@@ -13,6 +13,17 @@ would appreciate credit if this program or parts of it are used.
 #include <sys/types.h>
 #include <sys/ioctl.h>
 
+#ifdef TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
 #ifdef CRAY
 # ifndef TCSETCTTY
 #  if defined(HAVE_TERMIOS)
@@ -40,24 +51,16 @@ would appreciate credit if this program or parts of it are used.
 /*#include <memory.h> - deprecated - ANSI C moves them into string.h */
 #include "string.h"
 
-#ifdef HAVE_VARARGS_H
-#  include <varargs.h>
-#else
-#  ifdef HAVE_STDARG_H
-#    include <stdarg.h>
-#  endif
-#endif
-
 #include <errno.h>
 #include "exp_rename.h"
-#define EXP_DEFINE_FNS
+#define EXP_AVOID_INCLUDING_TCL_H
 #include "expect.h"
 #include "exp_int.h"
 
 #include "exp_printify.h"
 
 #ifdef NO_STDLIB_H
-#include "compat/stdlib.h"
+#include "../compat/stdlib.h"
 #else
 #include <stdlib.h>		/* for malloc */
 #endif
@@ -94,12 +97,20 @@ int Exp_StringMatch();
 
 void exp_init_pty();
 
-#ifndef HAVE_TCLREGERROR
-extern char *tclRegexpError;	/* declared in tclInt.h */
-#else
+/*
+   The following functions are linked from the Tcl library.  They
+   don't cause anything else in the library to be dragged in, so it
+   shouldn't cause any problems (e.g., bloat).
+
+   The functions are relatively small but painful enough that I don't care
+   to recode them.  You may, if you absolutely want to get rid of any
+   vestiges of Tcl.
+*/
 extern char *TclGetRegError();
 extern void TclRegError();
-#endif
+char *Tcl_ErrnoMsg();
+
+
 
 static unsigned int bufsiz = 2*EXP_MATCH_MAX;
 
@@ -265,7 +276,7 @@ when trapping, see below in child half of fork */
 		debuglog("parent: waiting for sync byte\r\n");
 		cc = read(sync_fds[0],&sync_byte,1);
 		if (cc == -1) {
-			fprintf(stderr,"parent sync byte read: %s\r\n",strerror(errno));
+			fprintf(stderr,"parent sync byte read: %s\r\n",Tcl_ErrnoMsg(errno));
 			exit(-1);
 		}
 
@@ -279,7 +290,7 @@ when trapping, see below in child half of fork */
 		debuglog("parent: telling child to go ahead\r\n");
 		cc = write(sync2_fds[1]," ",1);
 		if (cc == -1) {
-			errorlog("parent sync byte write: %s\r\n",strerror(errno));
+			errorlog("parent sync byte write: %s\r\n",Tcl_ErrnoMsg(errno));
 			exit(-1);
 		}
 
@@ -352,7 +363,7 @@ when trapping, see below in child half of fork */
 	    if (0 > (exp_pty[1] = getptyslave(exp_ttycopy,exp_ttyinit,
 						exp_stty_init))) {
 		restore_error_fd
-		fprintf(stderr,"open(slave pty): %s\n",strerror(errno));
+		fprintf(stderr,"open(slave pty): %s\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	    }
 	    /* sanity check */
@@ -380,7 +391,11 @@ when trapping, see below in child half of fork */
 #if defined(TIOCSCTTY) && !defined(sun) && !defined(hpux)
 	/* 4.3+BSD way to acquire controlling terminal */
 	/* according to Stevens - Adv. Prog..., p 642 */
+#ifdef __QNX__ /* posix in general */
+	if (tcsetct(0, getpid()) == -1) {
+#else
 	if (ioctl(0,TIOCSCTTY,(char *)0) < 0) {
+#endif
 		restore_error_fd
 		fprintf(stderr,"failed to get controlling terminal using TIOCSCTTY");
 		exit(-1);
@@ -393,7 +408,7 @@ when trapping, see below in child half of fork */
  	(void) close(0);
  	if (open("/dev/tty", O_RDWR) < 0) {
 		restore_error_fd
- 		fprintf(stderr,"open(/dev/tty): %s\r\n",strerror(errno));
+ 		fprintf(stderr,"open(/dev/tty): %s\r\n",Tcl_ErrnoMsg(errno));
  		exit(-1);
  	}
  	(void) close(1);
@@ -413,7 +428,7 @@ when trapping, see below in child half of fork */
 	 */
 	if ((pid = fork()) == -1) {
 		restore_error_fd
-		fprintf(stderr,"second fork: %s\r\n",strerror(errno));
+		fprintf(stderr,"second fork: %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 
@@ -476,7 +491,7 @@ when trapping, see below in child half of fork */
 	cc = write(sync_fds[1]," ",1);
 	if (cc == -1) {
 		restore_error_fd
-		fprintf(stderr,"child: sync byte write: %s\r\n",strerror(errno));
+		fprintf(stderr,"child: sync byte write: %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 	close(sync_fds[1]);
@@ -488,7 +503,7 @@ when trapping, see below in child half of fork */
 	cc = read(sync2_fds[0],&sync_byte,1);
 	if (cc == -1) {
 		restore_error_fd
-		errorlog("child: sync byte read: %s\r\n",strerror(errno));
+		errorlog("child: sync byte read: %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 	close(sync2_fds[0]);
@@ -504,14 +519,14 @@ when trapping, see below in child half of fork */
 	/* allow user to do anything else to child */
 	if (exp_child_exec_prelude) (*exp_child_exec_prelude)();
 
-        (void) execvp(argv[0],argv);
+        (void) execvp(file,argv);
 	/* Unfortunately, by now we've closed fd's to stderr, logfile and
 		debugfile.
 	   The only reasonable thing to do is to send back the error as
 	   part of the program output.  This will be picked up in an
 	   expect or interact command.
 	*/
-	fprintf(stderr,"execvp(%s): %s\n",file,strerror(errno));
+	fprintf(stderr,"execvp(%s): %s\n",file,Tcl_ErrnoMsg(errno));
 	exit(-1);
 	/*NOTREACHED*/
 }
@@ -519,15 +534,15 @@ when trapping, see below in child half of fork */
 /* returns fd of master side of pty */
 /*VARARGS*/
 int
-exp_spawnl EXP_VARARGS_DEF(char *,arg1)
+exp_spawnl TCL_VARARGS_DEF(char *,arg1)
 /*exp_spawnl(va_alist)*/
 /*va_dcl*/
 {
-	va_list args;
+	va_list args; /* problematic line here */
 	int i;
 	char *arg, **argv;
 
-	arg = EXP_VARARGS_START(char *,arg1,args);
+	arg = TCL_VARARGS_START(char *,arg1,args);
 	/*va_start(args);*/
 	for (i=1;;i++) {
 		arg = va_arg(args,char *);
@@ -536,7 +551,7 @@ exp_spawnl EXP_VARARGS_DEF(char *,arg1)
 	va_end(args);
 	if (i == 0) sysreturn(EINVAL);
 	if (!(argv = (char **)malloc((i+1)*sizeof(char *)))) sysreturn(ENOMEM);
-	argv[0] = EXP_VARARGS_START(char *,arg1,args);
+	argv[0] = TCL_VARARGS_START(char *,arg1,args);
 	/*va_start(args);*/
 	for (i=1;;i++) {
 		argv[i] = va_arg(args,char *);
@@ -625,9 +640,11 @@ int timeout;
 			int c;
 			c = getc(fp);
 			if (c == EOF) {
+/*fprintf(stderr,"<<EOF>>",c);fflush(stderr);*/
 				if (feof(fp)) cc = 0;
 				else cc = -1;
 			} else {
+/*fprintf(stderr,"<<%c>>",c);fflush(stderr);*/
 				buffer[0] = c;
 				cc = 1;
 			}
@@ -748,17 +765,9 @@ struct exp_case *ecases;
 	/* compile if necessary */
 	for (ec=ecases;ec->type != exp_end;ec++) {
 		if ((ec->type == exp_regexp) && !ec->re) {
-#ifndef HAVE_TCLREGERROR
-			tclRegexpError = 0;
-#else
 			TclRegError((char *)0);
-#endif
 			if (!(ec->re = TclRegComp(ec->pattern))) {
-#ifndef HAVE_TCLREGERROR
-				fprintf(stderr,"regular expression %s is bad: %s",ec->pattern,tclRegexpError);
-#else
 				fprintf(stderr,"regular expression %s is bad: %s",ec->pattern,TclGetRegError());
-#endif
 				return_errno(EINVAL);
 			  }
 		  }
@@ -783,13 +792,17 @@ struct exp_case *ecases;
 				bufsiz-1);
 			buf_length = bufsiz-1;
 		}
-		exp_buffer = ckrealloc(exp_buffer,bufsiz+1);
+		exp_buffer = realloc(exp_buffer,bufsiz+1);
 		if (!exp_buffer) return_errno(ENOMEM);
 		exp_buffer[buf_length] = '\0';
 		exp_buffer_end = exp_buffer + buf_length;
 		f->msize = bufsiz;
 	}
 
+	/* some systems (i.e., Solaris) require fp be flushed when switching */
+	/* directions - do this again afterwards */
+	if (fd == -1) fflush(fp);
+
 	if (exp_timeout != -1) signal(SIGALRM,sigalarm_handler);
 
 	/* remtime and current_time updated at bottom of loop */
@@ -830,7 +843,7 @@ struct exp_case *ecases;
 		/*
 		 * check for timeout
 		 */
-		if (remtime < 0 || polled) {
+		if ((exp_timeout >= 0) && ((remtime < 0) || polled)) {
 			debuglog("expect: timeout\r\n");
 			exp_match_end = exp_buffer;
 			return_normally(EXP_TIMEOUT);
@@ -928,22 +941,13 @@ struct exp_case *ecases;
 					}
 				}
 			} else {
-#ifndef HAVE_TCLREGERROR
-				tclRegexpError = 0;
-#else
 				TclRegError((char *)0);
-#endif
 				if (TclRegExec(ec->re,exp_buffer,exp_buffer)) {
 					matched = 1;
 					exp_match = ec->re->startp[0];
 					exp_match_end = ec->re->endp[0];
-#ifndef HAVE_TCLREGERROR
-				} else if (tclRegexpError) {
-			    		fprintf(stderr,"r.e. match (pattern %s) failed: %s",ec->pattern,tclRegexpError);
-#else
 				} else if (TclGetRegError()) {
 			    		fprintf(stderr,"r.e. match (pattern %s) failed: %s",ec->pattern,TclGetRegError());
-#endif
 				}
 			}
 
@@ -967,6 +971,11 @@ struct exp_case *ecases;
 	f->buffer     = exp_buffer;
 	f->buffer_end = exp_buffer_end;
 	f->match_end  = exp_match_end;
+
+	/* some systems (i.e., Solaris) require fp be flushed when switching */
+	/* directions - do this before as well */
+	if (fd == -1) fflush(fp);
+
 	if (sys_error) {
 		errno = sys_error;
 		return -1;
@@ -992,7 +1001,7 @@ struct exp_case *ecases;
 
 /*VARARGS*/
 int
-exp_expectl EXP_VARARGS_DEF(int,arg1)
+exp_expectl TCL_VARARGS_DEF(int,arg1)
 /*exp_expectl(va_alist)*/
 /*va_dcl*/
 {
@@ -1002,7 +1011,7 @@ exp_expectl EXP_VARARGS_DEF(int,arg1)
 	int i;
 	enum exp_type type;
 
-	fd = EXP_VARARGS_START(int,arg1,args);
+	fd = TCL_VARARGS_START(int,arg1,args);
 	/* va_start(args);*/
 	/* fd = va_arg(args,int);*/
 	/* first just count the arg sets */
@@ -1029,7 +1038,7 @@ exp_expectl EXP_VARARGS_DEF(int,arg1)
 		sysreturn(ENOMEM);
 
 	/* now set up the actual cases */
-	fd = EXP_VARARGS_START(int,arg1,args);
+	fd = TCL_VARARGS_START(int,arg1,args);
 	/*va_start(args);*/
 	/*va_arg(args,int);*/		/*COMPUTED BUT NOT USED*/
 	for (ec=ecases;;ec++) {
@@ -1057,7 +1066,7 @@ exp_expectl EXP_VARARGS_DEF(int,arg1)
 }
 
 int
-exp_fexpectl EXP_VARARGS_DEF(FILE *,arg1)
+exp_fexpectl TCL_VARARGS_DEF(FILE *,arg1)
 /*exp_fexpectl(va_alist)*/
 /*va_dcl*/
 {
@@ -1067,7 +1076,7 @@ exp_fexpectl EXP_VARARGS_DEF(FILE *,arg1)
 	int i;
 	enum exp_type type;
 
-	fp = EXP_VARARGS_START(FILE *,arg1,args);
+	fp = TCL_VARARGS_START(FILE *,arg1,args);
 	/*va_start(args);*/
 	/*fp = va_arg(args,FILE *);*/
 	/* first just count the arg-pairs */
@@ -1093,8 +1102,12 @@ exp_fexpectl EXP_VARARGS_DEF(FILE *,arg1)
 					malloc((1+i)*sizeof(struct exp_case))))
 		sysreturn(ENOMEM);
 
+#if 0
 	va_start(args);
 	va_arg(args,FILE *);		/*COMPUTED, BUT NOT USED*/
+#endif
+	(void) TCL_VARARGS_START(FILE *,arg1,args);
+
 	for (ec=ecases;;ec++) {
 		ec->type = va_arg(args,enum exp_type);
 		if (ec->type == exp_end) break;
