@@ -63,6 +63,10 @@ would appreciate credit if this program or parts of it are used.
 # endif
 #endif
 
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+
 #include <math.h>		/* for log/pow computation in send -h */
 #include <ctype.h>		/* all this for ispunct! */
 
@@ -148,7 +152,7 @@ RETSIGTYPE (*traps[])();
 /* Do not terminate format strings with \n!!! */
 /*VARARGS*/
 void
-exp_error EXP_VARARGS_DEF(Tcl_Interp *,arg1)
+exp_error TCL_VARARGS_DEF(Tcl_Interp *,arg1)
 /*exp_error(va_alist)*/
 /*va_dcl*/
 {
@@ -156,7 +160,7 @@ exp_error EXP_VARARGS_DEF(Tcl_Interp *,arg1)
 	char *fmt;
 	va_list args;
 
-	interp = EXP_VARARGS_START(Tcl_Interp *,arg1,args);
+	interp = TCL_VARARGS_START(Tcl_Interp *,arg1,args);
 	/*va_start(args);*/
 	/*interp = va_arg(args,Tcl_Interp *);*/
 	fmt = va_arg(args,char *);
@@ -208,7 +212,7 @@ int pid;
 static char close_cmd[] = "close";
 
 /* zero out the wait status field */
-static 
+static void
 exp_wait_zero(status)
 WAIT_STATUS_TYPE *status;
 {
@@ -329,21 +333,46 @@ close_tcl_file(interp,file_id)
 Tcl_Interp *interp;
 char *file_id;
 {
-	char *argv[3];
-	Tcl_CmdInfo info;
+    Tcl_VarEval(interp,"close ",file_id,(char *)0);
 
-	argv[0] = close_cmd;
-	argv[1] = file_id;
-	argv[2] = 0;
+#if 0  /* old Tcl 7.6 code */
+    char *argv[3];
+    Tcl_CmdInfo info;
 
-	Tcl_ResetResult(interp);
-	Tcl_GetCommandInfo(interp,"close",&info);
-	if (0 == Tcl_GetCommandInfo(interp,"close",&info)) {
-		info.clientData = 0;
-	}
-	(void) Tcl_CloseCmd(info.clientData,interp,2,argv);
+    argv[0] = close_cmd;
+    argv[1] = file_id;
+    argv[2] = 0;
+
+    Tcl_ResetResult(interp);
+    Tcl_GetCommandInfo(interp,"close",&info);
+    if (0 == Tcl_GetCommandInfo(interp,"close",&info)) {
+        info.clientData = 0;
+    }
+    (void) Tcl_CloseCmd(info.clientData,interp,2,argv);
+#endif
 }			
 
+
+/* close all connections
+The kernel would actually do this by default, however Tcl is going to
+come along later and try to reap its exec'd processes.  If we have
+inherited any via spawn -open, Tcl can hang if we don't close the
+connections first.
+*/
+
+void
+exp_close_all(interp)
+Tcl_Interp *interp;
+{
+	int fd;
+
+	for (fd=0;fd<=exp_fd_max;fd++) {
+		if (exp_fs[fd].valid) {
+			exp_close(interp,fd);
+		}
+	}
+}
+
 int
 exp_close(interp,fd)
 Tcl_Interp *interp;
@@ -415,7 +444,9 @@ int pid;
 			exp_fs[i].valid = FALSE;
 			exp_fs[i].fd_ptr = (int *)ckalloc(sizeof(int));
 			*exp_fs[i].fd_ptr = i;
+
 /*			exp_fs[i].ptr = (struct exp_f **)ckalloc(sizeof(struct exp_fs *));*/
+
 		}
 
 #if 0
@@ -440,6 +471,14 @@ int pid;
 	exp_fs[fd].key = expect_key++;
 	exp_fs[fd].force_read = FALSE;
 	exp_fs[fd].fg_armed = FALSE;
+#if TCL_MAJOR_VERSION < 8
+	/* Master must be inited each time because Tcl could have alloc'd */
+	/* this fd and shut it down (deallocating the FileHandle) behind */
+	/* our backs */
+	exp_fs[fd].Master = Tcl_GetFile((ClientData)fd,TCL_UNIX_FD);
+	exp_fs[fd].MasterOutput = 0;
+	exp_fs[fd].Slave = 0;
+#endif /* TCL_MAJOR_VERSION < 8 */
 	exp_fs[fd].tcl_handle = 0;
 	exp_fs[fd].slave_fd = EXP_NOFD;
 #ifdef HAVE_PTYTRAP
@@ -622,6 +661,8 @@ char **argv;
 
 	int sync_fds[2];
 	int sync2_fds[2];
+	int status_pipe[2];
+	int child_errno;
 	char sync_byte;
 
 	char buf[4];		/* enough space for a string literal */
@@ -653,9 +694,17 @@ char **argv;
 		} else if (streq(*argv,"-pty")) {
 			pty_only = TRUE;
 		} else if (streq(*argv,"-open")) {
+			if (argc < 2) {
+				exp_error(interp,"usage: -open file-identifier");
+				return TCL_ERROR;
+			}
 			openarg = argv[1];
 			argc--; argv++;
 		} else if (streq(*argv,"-leaveopen")) {
+			if (argc < 2) {
+				exp_error(interp,"usage: -open file-identifier");
+				return TCL_ERROR;
+			}
 			openarg = argv[1];
 			leaveopen = TRUE;
 			argc--; argv++;
@@ -663,7 +712,7 @@ char **argv;
 			int sig;
 
 			if (argc < 2) {
-				exp_error(interp,"usage: -trap siglist SIG_DFL or SIG_IGN");
+				exp_error(interp,"usage: -ignore signal");
 				return TCL_ERROR;
 			}
 			sig = exp_string_to_signal(interp,argv[1]);
@@ -768,7 +817,7 @@ when trapping, see below in child half of fork */
 			}
 
 			count = 0;
-			for (i=3;i<exp_fd_max;i++) {
+			for (i=3;i<=exp_fd_max;i++) {
 				count += exp_fs[i].valid;
 			}
 			if (count > 10) {
@@ -793,8 +842,16 @@ when trapping, see below in child half of fork */
 #define SPAWN_OUT "spawn_out"
 		Tcl_SetVar2(interp,SPAWN_OUT,"slave,name",exp_pty_slave_name,0);
 	} else {
+		Tcl_Channel chan;
+		int mode;
+#if TCL_MAJOR_VERSION < 8
+		Tcl_File tclReadFile, tclWriteFile;
+#endif /* TCL_MAJOR_VERSION < 8 */
+		int rfd, wfd;
+
 		if (echo) exp_log(0,"%s [open ...]\r\n",argv0);
 
+#if TCL7_4
 		rc = Tcl_GetOpenFile(interp,openarg,0,1,&readfilePtr);
 		wc = Tcl_GetOpenFile(interp,openarg,1,1,&writefilePtr);
 
@@ -820,6 +877,52 @@ when trapping, see below in child half of fork */
 			exp_close_on_exec(write_master);
 		}
 
+#endif
+		if (!(chan = Tcl_GetChannel(interp,openarg,&mode))) {
+			return TCL_ERROR;
+		}
+		if (!mode) {
+			exp_error(interp,"channel is neither readable nor writable");
+			return TCL_ERROR;
+		}
+		if (mode & TCL_READABLE) {
+#if TCL_MAJOR_VERSION < 8
+			tclReadFile = Tcl_GetChannelFile(chan, TCL_READABLE);
+			rfd = (int)Tcl_GetFileInfo(tclReadFile, (int *)0);
+#else
+			if (TCL_ERROR == Tcl_GetChannelHandle(chan, TCL_READABLE, (ClientData) &rfd)) {
+				return TCL_ERROR;
+			}
+#endif
+		}
+		if (mode & TCL_WRITABLE) {
+#if TCL_MAJOR_VERSION < 8
+			tclWriteFile = Tcl_GetChannelFile(chan, TCL_WRITABLE);
+			wfd = (int)Tcl_GetFileInfo(tclWriteFile, (int *)0);
+#else
+			if (TCL_ERROR == Tcl_GetChannelHandle(chan, TCL_WRITABLE, (ClientData) &wfd)) {
+				return TCL_ERROR;
+			}
+#endif
+		}
+
+		master = ((mode & TCL_READABLE)?rfd:wfd);
+
+		/* make a new copy of file descriptor */
+		if (-1 == (write_master = master = dup(master))) {
+			exp_error(interp,"fdopen: %s",Tcl_PosixError(interp));
+			return TCL_ERROR;
+		}
+
+		/* if writefilePtr is different, dup that too */
+		if ((mode & TCL_READABLE) && (mode & TCL_WRITABLE) && (wfd != rfd)) {
+			if (-1 == (write_master = dup(wfd))) {
+				exp_error(interp,"fdopen: %s",Tcl_PosixError(interp));
+				return TCL_ERROR;
+			}
+			exp_close_on_exec(write_master);
+		}
+
 		/*
 		 * It would be convenient now to tell Tcl to close its
 		 * file descriptor.  Alas, if involved in a pipeline, Tcl
@@ -897,7 +1000,7 @@ when trapping, see below in child half of fork */
 
 			if (0 > (f->slave_fd = getptyslave(ttycopy,ttyinit,
 					stty_init))) {
-				exp_error(interp,"open(slave pty): %s\r\n",strerror(errno));
+				exp_error(interp,"open(slave pty): %s\r\n",Tcl_PosixError(interp));
 				return TCL_ERROR;
 			}
 
@@ -931,6 +1034,13 @@ when trapping, see below in child half of fork */
 		goto parent_error;
 	}
 
+	if (-1 == pipe(status_pipe)) {
+		close(sync_fds[0]);
+		close(sync_fds[1]);
+		close(sync2_fds[0]);
+		close(sync2_fds[1]);
+	}
+
 	if ((pid = fork()) == -1) {
 		exp_error(interp,"fork: %s",Tcl_PosixError(interp));
 		goto parent_error;
@@ -941,6 +1051,7 @@ when trapping, see below in child half of fork */
 
 		close(sync_fds[1]);
 		close(sync2_fds[0]);
+		close(status_pipe[1]);
 
 		f = fd_new(master,pid);
 
@@ -1002,7 +1113,7 @@ when trapping, see below in child half of fork */
 			/* empty */;
 		}
 		if (rc == -1) {
-			errorlog("parent sync byte read: %s\r\n",strerror(errno));
+			errorlog("parent sync byte read: %s\r\n",Tcl_ErrnoMsg(errno));
 			exit(-1);
 		}
 
@@ -1016,15 +1127,40 @@ when trapping, see below in child half of fork */
 		debuglog("parent: telling child to go ahead\r\n");
 		wc = write(sync2_fds[1]," ",1);
 		if (wc == -1) {
-			errorlog("parent sync byte write: %s\r\n",strerror(errno));
+			errorlog("parent sync byte write: %s\r\n",Tcl_ErrnoMsg(errno));
 			exit(-1);
 		}
 
 		debuglog("parent: now unsynchronized from child\r\n");
-
 		close(sync_fds[0]);
 		close(sync2_fds[1]);
 
+		/* see if child's exec worked */
+	retry:
+		switch (read(status_pipe[0],&child_errno,sizeof child_errno)) {
+		case -1:
+			if (errno == EINTR) goto retry;
+			/* well it's not really the child's errno */
+			/* but it can be treated that way */
+			child_errno = errno;
+			break;
+		case 0:
+			/* child's exec succeeded */
+			child_errno = 0;
+			break;
+		default:
+			/* child's exec failed; err contains exec's errno  */
+			waitpid(pid, NULL, 0);
+			/* in order to get Tcl to set errorcode, we must */
+			/* hand set errno */
+			errno = child_errno;
+			exp_error(interp, "couldn't execute \"%s\": %s",
+				argv[0],Tcl_PosixError(interp));
+			goto parent_error;
+		}
+		close(status_pipe[0]);
+
+
 		/* tell user id of new process */
 		sprintf(buf,"%d",master);
 		Tcl_SetVar(interp,SPAWN_ID_VARNAME,buf,0);
@@ -1043,6 +1179,8 @@ parent_error:
 
 	close(sync_fds[0]);
 	close(sync2_fds[1]);
+	close(status_pipe[0]);
+	exp_close_on_exec(status_pipe[1]);
 
 	if (exp_dev_tty != -1) {
 		close(exp_dev_tty);
@@ -1109,7 +1247,7 @@ parent_error:
 
 	if (0 > (slave = getptyslave(ttycopy,ttyinit,stty_init))) {
 		restore_error_fd
-		errorlog("open(slave pty): %s\r\n",strerror(errno));
+		errorlog("open(slave pty): %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 	/* sanity check */
@@ -1129,7 +1267,11 @@ parent_error:
 	/* according to Stevens - Adv. Prog..., p 642 */
 	/* Oops, it appears that the CIBAUD is on Linux also */
 	/* so let's try without... */
+#ifdef __QNX__
+	if (tcsetct(0, getpid()) == -1) {
+#else
 	if (ioctl(0,TIOCSCTTY,(char *)0) < 0) {
+#endif
 		restore_error_fd
 		errorlog("failed to get controlling terminal using TIOCSCTTY");
 		exit(-1);
@@ -1142,7 +1284,7 @@ parent_error:
  	(void) close(0);
  	if (open("/dev/tty", O_RDWR) < 0) {
 		restore_error_fd
- 		errorlog("open(/dev/tty): %s\r\n",strerror(errno));
+ 		errorlog("open(/dev/tty): %s\r\n",Tcl_ErrnoMsg(errno));
  		exit(-1);
  	}
  	(void) close(1);
@@ -1162,7 +1304,7 @@ parent_error:
 	 */
 	if ((pid = fork()) == -1) {
 		restore_error_fd
-		errorlog("second fork: %s\r\n",strerror(errno));
+		errorlog("second fork: %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 
@@ -1229,7 +1371,7 @@ parent_error:
 	wc = write(sync_fds[1]," ",1);
 	if (wc == -1) {
 		restore_error_fd
-		errorlog("child: sync byte write: %s\r\n",strerror(errno));
+		errorlog("child: sync byte write: %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 	close(sync_fds[1]);
@@ -1245,7 +1387,7 @@ parent_error:
 
 	if (rc == -1) {
 		restore_error_fd
-		errorlog("child: sync byte read: %s\r\n",strerror(errno));
+		errorlog("child: sync byte read: %s\r\n",Tcl_ErrnoMsg(errno));
 		exit(-1);
 	}
 	close(sync2_fds[0]);
@@ -1257,20 +1399,17 @@ parent_error:
 	if (exp_close_in_child) (*exp_close_in_child)();
 
         (void) execvp(argv[0],argv);
+#if 0
 	/* Unfortunately, by now we've closed fd's to stderr, logfile and
 		debugfile.
 	   The only reasonable thing to do is to send back the error as
 	   part of the program output.  This will be picked up in an
 	   expect or interact command.
 	*/
-#if 0
-{ FILE *fp = fopen("/tmp/error","w");
-  fprintf(fp,"error occurred in execl\n");
-  fprintf(fp,"%s: %s\r\n",argv[0],strerror(errno));
-  fclose(fp);
-}
+	errorlog("%s: %s\r\n",argv[0],Tcl_ErrnoMsg(errno));
 #endif
-	errorlog("%s: %s\r\n",argv[0],strerror(errno));
+	/* if exec failed, communicate the reason back to the parent */
+	write(status_pipe[1], &errno, sizeof errno);
 	exit(-1);
 	/*NOTREACHED*/
 }
@@ -1367,7 +1506,15 @@ int rembytes;
 
 	while (rembytes) {
 		if (-1 == (cc = write(fd,buffer,rembytes))) return(-1);
-		/* if (0 == cc) return(-1); can't happen! */
+		if (0 == cc) {
+			/* This shouldn't happen but I'm told that it does */
+			/* nonetheless (at least on SunOS 4.1.3).  Since */
+			/* this is not a documented return value, the most */
+			/* reasonable thing is to complain here and retry */
+			/* in the hopes that is some transient condition. */
+			sleep(1);
+			exp_debuglog("write() failed to write anything but returned - sleeping and retrying...\n");
+		}
 
 		buffer += cc;
 		rembytes -= cc;
@@ -1386,8 +1533,6 @@ get_slow_args(interp,x)
 Tcl_Interp *interp;
 struct slow_arg *x;
 {
-	float ftime;
-
 	int sc;		/* return from scanf */
 	char *s = exp_get_var(interp,"send_slow");
 	if (!s) {
@@ -1607,6 +1752,7 @@ int val;
 	fd = exp_fd_list_pool;
 	exp_fd_list_pool = exp_fd_list_pool->next;
 	fd->fd = val;
+	/* fd->next is assumed to be changed by caller */
 	return fd;
 }
 
@@ -1781,7 +1927,7 @@ struct exp_i *i;
 		p = Tcl_GetVar(interp,i->variable,TCL_GLOBAL_ONLY);
 		if (!p) {
 			p = "";
-			exp_debuglog(interp,"warning: indirect variable %s undefined",i->variable);
+			exp_debuglog("warning: indirect variable %s undefined",i->variable);
 		}
 
 		if (i->value) {
@@ -2058,7 +2204,6 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
-	static char *logfile_tcl = 0;	/* logfile from Tcl's open */
 	static Tcl_DString dstring;
 	static int first_time = TRUE;
 	static int current_append;	/* true if currently appending */
@@ -2173,9 +2318,15 @@ char **argv;
 	} else if (openarg) {
 		int cc;
 		int fd;
+		Tcl_Channel chan;
+		int mode;
+#if TCL_MAJOR_VERSION < 8
+		Tcl_File tclWriteFile;
+#endif /* TCL_MAJOR_VERSION < 8 */
 
 		Tcl_DStringTrunc(&dstring,0);
 
+#if TCL7_4
 		cc = Tcl_GetOpenFile(interp,openarg,1,1,&writefilePtr);
 		if (cc == TCL_ERROR) goto error;
 
@@ -2183,6 +2334,23 @@ char **argv;
 			exp_error(interp,"dup: %s",Tcl_PosixError(interp));
 			goto error;
 		}
+#endif
+		if (!(chan = Tcl_GetChannel(interp,openarg,&mode))) {
+			return TCL_ERROR;
+		}
+		if (!(mode & TCL_WRITABLE)) {
+			exp_error(interp,"channel is not writable");
+		}
+#if TCL_MAJOR_VERSION < 8
+		tclWriteFile = Tcl_GetChannelFile(chan, TCL_WRITABLE);
+		fd = dup((int)Tcl_GetFileInfo(tclWriteFile, (int *)0));
+#else
+		if (TCL_ERROR == Tcl_GetChannelHandle(chan, TCL_WRITABLE, (ClientData) &fd)) {
+			goto error;
+		}
+		fd = dup(fd);
+#endif
+	
 		if (!(logfile = fdopen(fd,type))) {
 			exp_error(interp,"fdopen: %s",Tcl_PosixError(interp));
 			close(fd);
@@ -2425,7 +2593,7 @@ char **argv;
 			return TCL_OK;
 		} else if (exp_flageq(*argv,"-noexit",3)) {
 			argv++;
-			exp_exit_handlers(interp);
+			exp_exit_handlers((ClientData)interp);
 			return TCL_OK;
 		}
 	}
@@ -2440,45 +2608,59 @@ char **argv;
 	/*NOTREACHED*/
 }
 
+/* so cmd table later is more intuitive */
+#define Exp_CloseObjCmd Exp_CloseCmd
+
 /*ARGSUSED*/
 static int
 Exp_CloseCmd(clientData, interp, argc, argv)
 ClientData clientData;
 Tcl_Interp *interp;
 int argc;
+#if TCL_MAJOR_VERSION < 8
 char **argv;
+#else
+Tcl_Obj *CONST argv[];	/* Argument objects. */
+#endif
 {
 	int onexec_flag = FALSE;	/* true if -onexec seen */
 	int close_onexec;
 	int slave_flag = FALSE;
-#if 0
-	int slave;
-#endif
 	int m = -1;
 
 	int argc_orig = argc;
+#if TCL_MAJOR_VERSION < 8
 	char **argv_orig = argv;
+#else
+	Tcl_Obj *CONST *argv_orig = argv;
+#endif
 
 	argc--; argv++;
 
+#if TCL_MAJOR_VERSION < 8
+#define STARARGV *argv
+#else
+#define STARARGV Tcl_GetStringFromObj(*argv,(int *)0)
+#endif
+
 	for (;argc>0;argc--,argv++) {
-		if (streq(*argv,"-i")) {
+		if (streq("-i",STARARGV)) {
 			argc--; argv++;
-			if (!*argv) {
+			if (argc == 0) {
 				exp_error(interp,"usage: -i spawn_id");
 				return(TCL_ERROR);
 			}
-			m = atoi(*argv);
-		} else if (streq(*argv,"-slave")) {
+			m = atoi(STARARGV);
+		} else if (streq(STARARGV,"-slave")) {
 			slave_flag = TRUE;
-		} else if (streq(*argv,"-onexec")) {
+		} else if (streq(STARARGV,"-onexec")) {
 			argc--; argv++;
-			if (!*argv) {
+			if (argc == 0) {
 				exp_error(interp,"usage: -onexec 0|1");
 				return(TCL_ERROR);
 			}
 			onexec_flag = TRUE;
-			close_onexec = atoi(*argv);
+			close_onexec = atoi(STARARGV);
 		} else break;
 	}
 
@@ -2493,7 +2675,11 @@ char **argv;
 		if (0 == Tcl_GetCommandInfo(interp,"close",&info)) {
 			info.clientData = 0;
 		}
+#if TCL_MAJOR_VERSION < 8
 		return(Tcl_CloseCmd(info.clientData,interp,argc_orig,argv_orig));
+#else
+		return(Tcl_CloseObjCmd(info.clientData,interp,argc_orig,argv_orig));
+#endif
 	}
 
 	if (m == -1) {
@@ -2756,11 +2942,15 @@ char **argv;
 
 				/* pass to Tcl, so it can do wait */
 				/* in background */
+#if TCL_MAJOR_VERSION < 8
 				Tcl_DetachPids(1,&f->pid);
-				exp_wait_zero(&f->pid);
+#else
+				Tcl_DetachPids(1,(Tcl_Pid *)&f->pid);
+#endif
+				exp_wait_zero(&f->wait);
 			} else {
 				while (1) {
-					if (tcl_AsyncReady) {
+					if (Tcl_AsyncReady()) {
 						int rc = Tcl_AsyncInvoke(interp,TCL_OK);
 						if (rc != TCL_OK) return(rc);
 					}
@@ -2841,7 +3031,7 @@ char **argv;
 
 	if (result == -1) {
 		sprintf(interp->result,"%d %d -1 %d POSIX %s %s",
-			f->pid,m,errno,Tcl_ErrnoId(),strerror(errno));
+			f->pid,m,errno,Tcl_ErrnoId(),Tcl_ErrnoMsg(errno));
 		result = TCL_OK;
 	} else if (result == NO_CHILD) {
 		interp->result = "no children";
@@ -3120,26 +3310,6 @@ char **argv;
 	/* to return TCL_OK, type "return" */
 }
 
-/* this command supercede's Tcl's builtin CONTINUE command */
-/*ARGSUSED*/
-int
-Exp_ExpContinueDeprecatedCmd(clientData, interp, argc, argv)
-ClientData clientData;
-Tcl_Interp *interp;
-int argc;
-char **argv;
-{
-	if (argc == 1) return(TCL_CONTINUE);
-	else if (argc == 2) {
-		if (streq(argv[1],"-expect")) {
-			debuglog("continue -expect is deprecated, use exp_continue\r\n");
-			return(EXP_CONTINUE);
-		}
-	}
-	exp_error(interp,"usage: continue [-expect]\n");
-	return(TCL_ERROR);
-}
-
 /* this command supercede's Tcl's builtin CONTINUE command */
 /*ARGSUSED*/
 int
@@ -3159,6 +3329,7 @@ char **argv;
 	return(TCL_ERROR);
 }
 
+#if TCL_MAJOR_VERSION < 8
 /* most of this is directly from Tcl's definition for return */
 /*ARGSUSED*/
 int
@@ -3177,6 +3348,26 @@ char **argv;
 		result = EXP_TCL_RETURN;
 	return result;
 }
+#else
+/* most of this is directly from Tcl's definition for return */
+/*ARGSUSED*/
+int
+Exp_InterReturnObjCmd(clientData, interp, objc, objv)
+ClientData clientData;
+Tcl_Interp *interp;
+int objc;
+Tcl_Obj *CONST objv[];
+{
+    /* let Tcl's return command worry about args */
+    /* if successful (i.e., TCL_RETURN is returned) */
+    /* modify the result, so that we will handle it specially */
+
+    int result = Tcl_ReturnObjCmd(clientData,interp,objc,objv);
+    if (result == TCL_RETURN)
+        result = EXP_TCL_RETURN;
+    return result;
+}
+#endif
 
 /*ARGSUSED*/
 int
@@ -3186,12 +3377,11 @@ Tcl_Interp *interp;
 int argc;
 char **argv;
 {
-	char *tcl_handle = 0;
-	FILE *fp;
 	struct exp_f *f;
 	int m = -1;
 	int m2;
 	int leaveopen = FALSE;
+	Tcl_Channel chan;
 
 	argc--; argv++;
 
@@ -3215,122 +3405,37 @@ char **argv;
 
 	if (0 == (f = exp_fd2f(interp,m,1,0,"exp_open"))) return TCL_ERROR;
 
-/* old code - this code used to remove spawn id from Expect's memory */
-/* in anticipation of passing to Tcl - now we let Expect remember */
-#if 0
-	/* remove from Expect's memory in anticipation of passing to Tcl */
-	if (f->pid != EXP_NOPID) {
-		Tcl_DetachPids(1,&f->pid);
-	}
-
-	if (f->tcl_handle) {
-		tcl_handle = f->tcl_handle;
-	} else {
-		/* User probably forgot to close this earlier. */
-		if (f->slave_fd) close(f->slave_fd);
-	}
-	exp_f_prep_for_invalidation(interp,f);
-	f->valid = FALSE;
-
-	if (tcl_handle) {
-		/* already allocated by Tcl */
-		Tcl_SetResult(interp,tcl_handle,TCL_DYNAMIC);
-	} else {
-		if (!(fp = fdopen(m,"r+"))) {
-			exp_error(interp,"fdopen: %s",Tcl_PosixError(interp));
-			return TCL_ERROR;
-		}
-		Tcl_EnterFile(interp,fp,TCL_FILE_READABLE|TCL_FILE_WRITABLE);
-	}
-#endif
-
 	/* make a new copy of file descriptor */
 	if (-1 == (m2 = dup(m))) {
 		exp_error(interp,"fdopen: %s",Tcl_PosixError(interp));
 		return TCL_ERROR;
 	}
-	if (!(fp = fdopen(m2,"r+"))) {
-		exp_error(interp,"fork: %s",Tcl_PosixError(interp));
-		return TCL_ERROR;
-	}
 
 	if (!leaveopen) {
 		/* remove from Expect's memory in anticipation of passing to Tcl */
 		if (f->pid != EXP_NOPID) {
+#if TCL_MAJOR_VERSION < 8
 			Tcl_DetachPids(1,&f->pid);
+#else
+			Tcl_DetachPids(1,(Tcl_Pid *)&f->pid);
+#endif
 			f->pid = EXP_NOPID;
 			f->sys_waited = f->user_waited = TRUE;
 		}
 		exp_close(interp,m);
 	}
 
-	Tcl_EnterFile(interp,fp,TCL_FILE_READABLE|TCL_FILE_WRITABLE);
-
+	chan = Tcl_MakeFileChannel(
+#if TCL_MAJOR_VERSION < 8
+			    (ClientData)m2,
+#endif
+			    (ClientData)m2,
+			    TCL_READABLE|TCL_WRITABLE);
+	Tcl_RegisterChannel(interp, chan);
+	Tcl_AppendResult(interp, Tcl_GetChannelName(chan), (char *) NULL);
 	return TCL_OK;
 }
 
-#ifdef OLD_EXP_OPEN
-/*ARGSUSED*/
-int
-Exp_OpenCmd(clientData, interp, argc, argv)
-ClientData clientData;
-Tcl_Interp *interp;
-int argc;
-char **argv;
-{
-	char *tcl_handle = 0;
-	FILE *fp;
-	struct exp_f *f;
-	int m = -1;
-
-	argc--; argv++;
-
-	for (;argc>0;argc--,argv++) {
-		if (streq(*argv,"-i")) {
-			argc--; argv++;
-			if (!*argv) {
-				exp_error(interp,"usage: -i spawn_id");
-				return TCL_ERROR;
-			}
-			m = atoi(*argv);
-		} else break;
-	}
-
-	if (m == -1) {
-		if (exp_update_master(interp,&m,0,0) == 0) return TCL_ERROR;
-	}
-
-	if (0 == (f = exp_fd2f(interp,m,1,0,"exp_open"))) return TCL_ERROR;
-
-	/* remove from Expect's memory in anticipation of passing to Tcl */
-
-	if (f->pid != EXP_NOPID) {
-		Tcl_DetachPids(1,&f->pid);
-	}
-
-	if (f->tcl_handle) {
-		tcl_handle = f->tcl_handle;
-	} else {
-		/* User probably forgot to close this earlier. */
-		if (f->slave_fd) close(f->slave_fd);
-	}
-	exp_f_prep_for_invalidation(interp,f);
-	f->valid = FALSE;
-
-	if (tcl_handle) {
-		/* already allocated by Tcl */
-		Tcl_SetResult(interp,tcl_handle,TCL_DYNAMIC);
-	} else {
-		if (!(fp = fdopen(m,"r+"))) {
-			exp_error(interp,"fork: %s",Tcl_PosixError(interp));
-			return TCL_ERROR;
-		}
-		Tcl_EnterFile(interp,fp,TCL_FILE_READABLE|TCL_FILE_WRITABLE);
-	}
-	return TCL_OK;
-}
-#endif /*OLD_EXP_OPEN*/
-
 /* return 1 if a string is substring of a flag */
 /* this version is the code used by the macro that everyone calls */
 int
@@ -3352,10 +3457,16 @@ exp_create_commands(interp,c)
 Tcl_Interp *interp;
 struct exp_cmd_data *c;
 {
+#if TCL_MAJOR_VERSION < 8
 	Interp *iPtr = (Interp *) interp;
+#else
+	Namespace *globalNsPtr = (Namespace *) Tcl_GetGlobalNamespace(interp);
+	Namespace *currNsPtr   = (Namespace *) Tcl_GetCurrentNamespace(interp);
+#endif
 	char cmdnamebuf[80];
 
 	for (;c->name;c++) {
+#if TCL_MAJOR_VERSION < 8
 		int create = FALSE;
 		/* if already defined, don't redefine */
 		if (c->flags & EXP_REDEFINE) create = TRUE;
@@ -3366,46 +3477,74 @@ struct exp_cmd_data *c;
 			Tcl_CreateCommand(interp,c->name,c->proc,
 				c->data,exp_deleteProc);
 		}
+#else
+		/* if already defined, don't redefine */
+		if ((c->flags & EXP_REDEFINE) ||
+		    !(Tcl_FindHashEntry(&globalNsPtr->cmdTable,c->name) ||
+		      Tcl_FindHashEntry(&currNsPtr->cmdTable,c->name))) {
+			if (c->objproc)
+				Tcl_CreateObjCommand(interp,c->name,
+						     c->objproc,c->data,exp_deleteObjProc);
+			else
+				Tcl_CreateCommand(interp,c->name,c->proc,
+						  c->data,exp_deleteProc);
+		}
+#endif
 		if (!(c->name[0] == 'e' &&
 		      c->name[1] == 'x' &&
 		      c->name[2] == 'p')
 		    && !(c->flags & EXP_NOPREFIX)) {
 			sprintf(cmdnamebuf,"exp_%s",c->name);
+#if TCL_MAJOR_VERSION < 8
 			Tcl_CreateCommand(interp,cmdnamebuf,c->proc,
 				c->data,exp_deleteProc);
+#else
+			if (c->objproc)
+				Tcl_CreateObjCommand(interp,cmdnamebuf,c->objproc,c->data,
+						     exp_deleteObjProc);
+			else
+				Tcl_CreateCommand(interp,cmdnamebuf,c->proc,
+						  c->data,exp_deleteProc);
+#endif
 		}
 	}
 }
 
 static struct exp_cmd_data cmd_data[]  = {
+#if TCL_MAJOR_VERSION < 8
 {"close",	Exp_CloseCmd,	0,	EXP_REDEFINE},
-#ifdef TCL_DEBUGGER
-{"debug",	Exp_DebugCmd,	0,	0},
+#else
+{"close",	Exp_CloseObjCmd,	0,	0,	EXP_REDEFINE},
 #endif
-{"exp_internal",Exp_ExpInternalCmd,	0,	0},
-{"disconnect",	Exp_DisconnectCmd,	0,	0},
-{"exit",	Exp_ExitCmd,	0,	EXP_REDEFINE},
-{"continue",	Exp_ExpContinueDeprecatedCmd,0,EXP_NOPREFIX|EXP_REDEFINE},
-{"exp_continue",Exp_ExpContinueCmd,0,	0},
-{"fork",	Exp_ForkCmd,	0,	0},
-{"exp_pid",	Exp_ExpPidCmd,	0,	0},
-{"getpid",	Exp_GetpidDeprecatedCmd,0,	0},
-{"interpreter",	Exp_InterpreterCmd,	0,	0},
-{"log_file",	Exp_LogFileCmd,	0,	0},
-{"log_user",	Exp_LogUserCmd,	0,	0},
-{"exp_open",	Exp_OpenCmd,	0,	0},
-{"overlay",	Exp_OverlayCmd,	0,	0},
+#ifdef TCL_DEBUGGER
+{"debug",	exp_proc(Exp_DebugCmd),	0,	0},
+#endif
+{"exp_internal",exp_proc(Exp_ExpInternalCmd),	0,	0},
+{"disconnect",	exp_proc(Exp_DisconnectCmd),	0,	0},
+{"exit",	exp_proc(Exp_ExitCmd),	0,	EXP_REDEFINE},
+{"exp_continue",exp_proc(Exp_ExpContinueCmd),0,	0},
+{"fork",	exp_proc(Exp_ForkCmd),	0,	0},
+{"exp_pid",	exp_proc(Exp_ExpPidCmd),	0,	0},
+{"getpid",	exp_proc(Exp_GetpidDeprecatedCmd),0,	0},
+{"interpreter",	exp_proc(Exp_InterpreterCmd),	0,	0},
+{"log_file",	exp_proc(Exp_LogFileCmd),	0,	0},
+{"log_user",	exp_proc(Exp_LogUserCmd),	0,	0},
+{"exp_open",	exp_proc(Exp_OpenCmd),	0,	0},
+{"overlay",	exp_proc(Exp_OverlayCmd),	0,	0},
+#if TCL_MAJOR_VERSION < 8
 {"inter_return",Exp_InterReturnCmd,	0,	0},
-{"send",	Exp_SendCmd,	(ClientData)&sendCD_proc,	0},
-{"send_spawn",	Exp_SendCmd,	(ClientData)&sendCD_proc,	0},/*deprecat*/
-{"send_error",	Exp_SendCmd,	(ClientData)&sendCD_error,	0},
-{"send_log",	Exp_SendLogCmd,	0,	0},
-{"send_tty",	Exp_SendCmd,	(ClientData)&sendCD_tty,	0},
-{"send_user",	Exp_SendCmd,	(ClientData)&sendCD_user,	0},
-{"sleep",	Exp_SleepCmd,	0,	0},
-{"spawn",	Exp_SpawnCmd,	0,	0},
-{"strace",	Exp_StraceCmd,	0,	0},
-{"wait",	Exp_WaitCmd,	0,	0},
+#else
+{"inter_return",Exp_InterReturnObjCmd,	0,	0,	0},
+#endif
+{"send",	exp_proc(Exp_SendCmd),	(ClientData)&sendCD_proc,	0},
+{"send_error",	exp_proc(Exp_SendCmd),	(ClientData)&sendCD_error,	0},
+{"send_log",	exp_proc(Exp_SendLogCmd),	0,	0},
+{"send_tty",	exp_proc(Exp_SendCmd),	(ClientData)&sendCD_tty,	0},
+{"send_user",	exp_proc(Exp_SendCmd),	(ClientData)&sendCD_user,	0},
+{"sleep",	exp_proc(Exp_SleepCmd),	0,	0},
+{"spawn",	exp_proc(Exp_SpawnCmd),	0,	0},
+{"strace",	exp_proc(Exp_StraceCmd),	0,	0},
+{"wait",	exp_proc(Exp_WaitCmd),	0,	0},
 {0}};
 
 void
