@@ -7,9 +7,6 @@ dollars.  Therefore it is public domain.  However, the author and NIST
 would appreciate credit if this program or parts of it are used.
 */
 
-EXTERN void		exp_background_error_default
-				_ANSI_ARGS_((Tcl_Interp *));
-EXTERN void		exp_background_error _ANSI_ARGS_((Tcl_Interp *));
 EXTERN struct exp_f *	exp_update_master
 				_ANSI_ARGS_((Tcl_Interp *,int *,int,int));
 EXTERN char *		exp_get_var _ANSI_ARGS_((Tcl_Interp *,char *));
@@ -143,7 +140,7 @@ EXTERN int exp_flageq_code _ANSI_ARGS_((char *,char *,int));
 #define EXP_CMD_BEFORE	0
 #define EXP_CMD_AFTER	1
 #define EXP_CMD_BG	2
-#define EXP_CMD_UNUSED	-1
+#define EXP_CMD_FG	3
 
 /* each process is associated with a 'struct exp_f'.  An array of these */
 /* ('exp_fs') keeps track of all processes.  They are indexed by the true fd */
@@ -182,6 +179,21 @@ struct exp_f {
 			/* data).  This supports interact CAN_MATCH */
 	int fg_armed;	/* If Tk_CreateFileHandler is active for responding */
 			/* to foreground events */
+
+
+#if TCL_MAJOR_VERSION < 8
+	Tcl_File Master;	/* corresponds to master fd */
+	Tcl_File Slave;		/* corresponds to slave_fd */
+	Tcl_File MasterOutput;	/* corresponds to tcl_output */
+	/*
+	 *  Following comment only applies to Tcl 7.6:
+	 *  Explicit fds aren't necessary now, but since the code is already
+	 *  here from before Tcl required Tcl_File, we'll continue using
+	 *  the old fds.  If we ever port this code to a non-UNIX system,
+	 *  we'll dump the fds totally.
+	 */
+#endif /* TCL_MAJOR_VERSION < 8 */
+	   
 	int slave_fd;	/* slave fd if "spawn -pty" used */
 #ifdef HAVE_PTYTRAP
 	char *slave_name;/* Full name of slave, i.e., /dev/ttyp0 */
@@ -222,6 +234,7 @@ EXTERN struct exp_f *	exp_fd2f _ANSI_ARGS_((Tcl_Interp *,int,int,int,char *));
 EXTERN void		exp_adjust _ANSI_ARGS_((struct exp_f *));
 EXTERN void		exp_buffer_shuffle _ANSI_ARGS_((Tcl_Interp *,struct exp_f *,int,char *,char *));
 EXTERN int		exp_close _ANSI_ARGS_((Tcl_Interp *,int));
+EXTERN void		exp_close_all _ANSI_ARGS_((Tcl_Interp *));
 EXTERN void		exp_ecmd_remove_fd_direct_and_indirect 
 				_ANSI_ARGS_((Tcl_Interp *,int));
 EXTERN void		exp_trap_on _ANSI_ARGS_((int));
@@ -230,6 +243,7 @@ EXTERN int		exp_trap_off _ANSI_ARGS_((char *));
 EXTERN void		exp_strftime();
 
 #define exp_deleteProc (void (*)())0
+#define exp_deleteObjProc (void (*)())0
 
 EXTERN int expect_key;
 EXTERN int exp_configure_count;	/* # of times descriptors have been closed */
@@ -238,6 +252,7 @@ EXTERN int exp_nostack_dump;	/* TRUE if user has requested unrolling of */
 				/* stack with no trace */
 
 EXTERN void		exp_init_pty _ANSI_ARGS_((void));
+EXTERN void		exp_pty_exit _ANSI_ARGS_((void));
 EXTERN void		exp_init_tty _ANSI_ARGS_((void));
 EXTERN void		exp_init_stdio _ANSI_ARGS_((void));
 /*EXTERN void		exp_init_expect _ANSI_ARGS_((Tcl_Interp *));*/
@@ -305,8 +320,19 @@ EXTERN void		exp_i_update _ANSI_ARGS_((Tcl_Interp *,
 #define EXP_NOPREFIX	1	/* don't define with "exp_" prefix */
 #define EXP_REDEFINE	2	/* stomp on old commands with same name */
 
+/* a hack for easily supporting both Tcl 7 and 8 CreateCommand/Obj split */
+/* Can be discarded with Tcl 7 is. */
+#if TCL_MAJOR_VERSION < 8
+#define exp_proc(cmdproc) cmdproc
+#else
+#define exp_proc(cmdproc) 0, cmdproc
+#endif
+
 struct exp_cmd_data {
 	char		*name;
+#if TCL_MAJOR_VERSION >= 8
+	Tcl_ObjCmdProc	*objproc;
+#endif
 	Tcl_CmdProc	*proc;
 	ClientData	data;
 	int 		flags;
