@@ -6,17 +6,26 @@ Design and implementation of this program was paid for by U.S. tax
 dollars.  Therefore it is public domain.  However, the author and NIST
 would appreciate credit if this program or parts of it are used.
 
-$Revision: 1.2 $
-$Date: 1993/03/03 00:21:20 $
-
 */
 
 #include "expect_cf.h"
 #include <stdio.h>
 #include <sys/types.h>
-/*#include <sys/time.h> seems to not be present on SVR3 systems */
-/* and it's not used anyway as far as I can tell */
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
 
+#ifdef TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+  
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
@@ -36,12 +45,8 @@ $Date: 1993/03/03 00:21:20 $
 #include "tclRegexp.h"
 #include "exp_regexp.h"
 
-#ifndef HAVE_TCLREGERROR
-extern char *tclRegexpError;	/* declared in tclInt.h */
-#else
 extern char *TclGetRegError();
 extern void TclRegError();
-#endif
 
 #define INTER_OUT "interact_out"
 
@@ -895,18 +900,10 @@ char **argv;
 		km->null = FALSE;
 		km->re = 0;
 		if (next_re) {
-#ifndef HAVE_TCLREGERROR
-			tclRegexpError = 0;
-#else
 			TclRegError((char *)0);
-#endif
 			if (0 == (km->re = TclRegComp(*argv))) {
 				exp_error(interp,"bad regular expression: %s",
-#ifndef HAVE_TCLREGERROR
-								tclRegexpError);
-#else
 								TclGetRegError());
-#endif
 				return(TCL_ERROR);
 			}
 			next_re = FALSE;
@@ -1042,12 +1039,8 @@ char **argv;
 	}
 
 	for (inp = input_base,i=0;inp;inp=inp->next,i++) {
-	    struct exp_fd_list *fdp;
-
-	    for (fdp = inp->i_list->fd_list;fdp;fdp=fdp->next) {
-		/* start timers */
-		inp->timeout_remaining = inp->timeout_nominal;
-	    }
+	    /* start timers */
+	    inp->timeout_remaining = inp->timeout_nominal;
 	}
 
 	key = expect_key++;
@@ -2245,7 +2238,7 @@ struct output *o;
 }
 
 static struct exp_cmd_data cmd_data[]  = {
-{"interact",	Exp_InteractCmd,	0,	0},
+{"interact",	exp_proc(Exp_InteractCmd),	0,	0},
 {0}};
 
 void
