@@ -11,7 +11,6 @@ would appreciate credit if this program or parts of it are used.
 #include <sys/types.h>
 #include <stdio.h>
 #include <signal.h>
-#include <varargs.h>
 #include <errno.h>
 #include <ctype.h>	/* for isspace */
 #include <time.h>	/* for time(3) */
@@ -25,11 +24,12 @@ would appreciate credit if this program or parts of it are used.
 #include <sys/wait.h>
 #endif
 
-#include "tcl.h"
-#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION < 4
-extern char *tclRegexpError;	/* declared in tclInt.h */
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
 #endif
 
+#include "tcl.h"
+
 #include "string.h"
 
 #include "tclRegexp.h"
@@ -104,7 +104,11 @@ struct exp_cmd_descriptor {
 	int timeout;			/* timeout period if flag used */
 	struct exp_cases_descriptor ecd;
 	struct exp_i *i_list;
-} exp_cmds[3];
+} exp_cmds[4];
+/* note that exp_cmds[FG] is just a fake, the real contents is stored
+   in some dynamically-allocated variable.  We use exp_cmds[FG] mostly
+   as a well-known address and also as a convenience and so we allocate
+   just a few of its fields that we need. */
 
 static void
 exp_cmd_init(cmd,cmdtype,duration)
@@ -376,10 +380,10 @@ char **argv;
 		for (;isspace(*a);a++) {
 			if (*a == '\n') *a = ' ';
 		}
-#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION == 0
-		a = TclWordEnd(a,0)+1;
-#else
+#if TCL_MAJOR_VERSION < 8
 		a = TclWordEnd(a,0,(int *)0)+1;
+#else
+		a = TclWordEnd(a,&a[strlen(a)],0,(int *)0)+1;
 #endif
 	}
 
@@ -493,18 +497,10 @@ char **argv;
 			} else if (exp_flageq("regexp",arg,2)) {
 				i++;
 				ec.use = PAT_RE;
-#ifndef HAVE_TCLREGERROR
-				tclRegexpError = 0;
-#else
 				TclRegError((char *)0);
-#endif
 				if (!(ec.re = TclRegComp(argv[i]))) {
 					exp_error(interp,"bad regular expression: %s",
-#ifndef HAVE_TCLREGERROR
-								tclRegexpError);
-#else
 								TclGetRegError());
-#endif
 					goto error;
 				}
 			} else if (exp_flageq("exact",arg,2)) {
@@ -682,11 +678,7 @@ char *suffix;
 
 	if (e->use == PAT_RE) {
 		debuglog("\"%s\"? ",dprintify(e->pat));
-#ifndef HAVE_TCLREGERROR
-		tclRegexpError = 0;
-#else
 		TclRegError((char *)0);
-#endif
 		if (buffer && TclRegExec(e->re,buffer,buffer)) {
 			o->e = e;
 			o->match = e->re->endp[0]-buffer;
@@ -696,13 +688,8 @@ char *suffix;
 			return(EXP_MATCH);
 		} else {
 			debuglog(no);
-#ifndef HAVE_TCLREGERROR
-			if (tclRegexpError) {
-			    exp_error(interp,"-re failed: %s",tclRegexpError);
-#else
 			if (TclGetRegError()) {
 			    exp_error(interp,"-re failed: %s",TclGetRegError());
-#endif
 			    return(EXP_TCLERROR);
 		        }
 		    }
@@ -867,6 +854,10 @@ struct exp_i *exp_i;
 					sizeof(struct exp_cmd_descriptor *));
 			}
 			ecmd->ecd.count--;
+			if (0 == ecmd->ecd.count) {
+				ckfree((char *)ecmd->ecd.cases);
+				ecmd->ecd.cases = 0;
+			}
 		} else {
 			i++;
 		}
@@ -925,7 +916,7 @@ int direct;
 				exp_free_fd_single(tmp);
 
 				/* if last bg ecase, disarm spawn id */
-				if (ecmd->cmdtype == EXP_CMD_BG) {
+				if ((ecmd->cmdtype == EXP_CMD_BG) && (m != EXP_SPAWN_ID_ANY)) {
 					exp_fs[m].bg_ecount--;
 					if (exp_fs[m].bg_ecount == 0) {
 						exp_disarm_background_filehandler(m);
@@ -1198,6 +1189,7 @@ char **argv;
 	return TCL_OK;
 }
 
+/* Exp_ExpectGlobalCmd is invoked to process expect_before/after */
 /*ARGSUSED*/
 int
 Exp_ExpectGlobalCmd(clientData, interp, argc, argv)
@@ -1347,7 +1339,7 @@ char **argv;
 
 		if (ecmd->ecd.count) {
 			/* append to end */
-			ecmd->ecd.cases = (struct ecase **)ckrealloc(ecmd->ecd.cases, count * sizeof(struct ecase *));
+			ecmd->ecd.cases = (struct ecase **)ckrealloc((char *)ecmd->ecd.cases, count * sizeof(struct ecase *));
 			start_index = ecmd->ecd.count;
 		} else {
 			/* append to beginning */
@@ -1390,35 +1382,6 @@ char **argv;
 	return(result);
 }
 
-/* handle background errors - a skeletal version of Tk_BackgroundError */
-/* If we are in Tk, we skip this and use Tk's version */
-/* If Tcl ever gets a background error handler, we can dump this */
-void
-exp_background_error_default(interp)
-    Tcl_Interp *interp;		/* Interpreter in which an error has
-				 * occurred. */
-{
-    char *error;
-    char *tmp;
-
-    /*
-     * The Tcl_AddErrorInfo call below (with an empty string) ensures that
-     * errorInfo gets properly set.  It's needed in cases where the error
-     * came from a utility procedure like Tcl_GetVar instead of Tcl_Eval;
-     * in these cases errorInfo still won't have been set when this
-     * procedure is called.
-     */
-
-    Tcl_AddErrorInfo(interp, "");
-    error = (char *) ckalloc((unsigned) (strlen(interp->result) + 1));
-    strcpy(error, interp->result);
-    tmp = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
-    if (tmp == NULL) tmp = error;
-    fprintf(stderr, "%s\n", tmp);
-    Tcl_ResetResult(interp);
-    ckfree(error);
-}
-
 /* adjusts file according to user's size request */
 void
 exp_adjust(f)
@@ -1607,11 +1570,6 @@ int key;
 	return(cc);
 }
 
-/* this should really be local to i_read, however the longjmp could then */
-/* clobber them */
-static int i_read_cc;
-
-
 /* when buffer fills, copy second half over first and */
 /* continue, so we can do matches over multiple buffers */
 void
@@ -1731,12 +1689,12 @@ int save_flags;
 			if (alarm_fired) {
 				return EXP_TIMEOUT;
 			} else {
-				if (tcl_AsyncReady) {
+				if (Tcl_AsyncReady()) {
 					int rc = Tcl_AsyncInvoke(interp,TCL_OK);
 					if (rc != TCL_OK) return(exp_tcl2_returnvalue(rc));
 				}
 				if (!f->valid) {
-					exp_error("spawn_id %d no longer valid",f-exp_fs);
+					exp_error(interp,"spawn_id %d no longer valid",f-exp_fs);
 					return EXP_TCLERROR;
 				}
 				goto restart;
@@ -1816,7 +1774,7 @@ exp_cmdtype_printable(cmdtype)
 int cmdtype;
 {
 	switch (cmdtype) {
-	case EXP_CMD_UNUSED: return("expect");
+	case EXP_CMD_FG: return("expect");
 	case EXP_CMD_BG: return("expect_background");
 	case EXP_CMD_BEFORE: return("expect_before");
 	case EXP_CMD_AFTER: return("expect_after");
@@ -2003,7 +1961,7 @@ do_more_data:
 			m,&eo,&last_f,&last_case,cc,&m,1,"_background");
 	if (cc == EXP_TCLERROR) {
 		/* only likely problem here is some internal regexp botch */
-		exp_background_error(interp);
+		Tcl_BackgroundError(interp);
 		goto finish;
 	}
 	/* special eof code that cannot be done in eval_cases */
@@ -2186,7 +2144,7 @@ do_more_data:
 
 		if (body) {
 			int result = Tcl_GlobalEval(interp,body);
-			if (result != TCL_OK) exp_background_error(interp);
+			if (result != TCL_OK) Tcl_BackgroundError(interp);
 
 			if (eof_body) ckfree(eof_body);
 		}
@@ -2270,7 +2228,7 @@ char **argv;
 	/* make arg list for processing cases */
 	/* do it dynamically, since expect can be called recursively */
 
-	exp_cmd_init(&eg,EXP_CMD_UNUSED,EXP_TEMPORARY);
+	exp_cmd_init(&eg,EXP_CMD_FG,EXP_TEMPORARY);
 	fd_list = 0;
 	masters = 0;
 	if (TCL_ERROR == parse_expect_args(interp,&eg,
@@ -3014,16 +2972,16 @@ static int spawn_id_user = EXP_SPAWN_ID_USER;
 
 static struct exp_cmd_data
 cmd_data[]  = {
-{"expect",	Exp_ExpectCmd,	(ClientData)&spawn_id_bad,	0},
-{"expect_after",Exp_ExpectGlobalCmd,(ClientData)&exp_cmds[EXP_CMD_AFTER],0},
-{"expect_before",Exp_ExpectGlobalCmd,(ClientData)&exp_cmds[EXP_CMD_BEFORE],0},
-{"expect_user",	Exp_ExpectCmd,	(ClientData)&spawn_id_user,	0},
-{"expect_tty",	Exp_ExpectCmd,	(ClientData)&exp_dev_tty,	0},
-{"expect_background",Exp_ExpectGlobalCmd,(ClientData)&exp_cmds[EXP_CMD_BG],0},
-{"match_max",	Exp_MatchMaxCmd,	0,	0},
-{"remove_nulls",Exp_RemoveNullsCmd,	0,	0},
-{"parity",	Exp_ParityCmd,		0,	0},
-{"timestamp",	Exp_TimestampCmd,	0,	0},
+{"expect",	exp_proc(Exp_ExpectCmd),	(ClientData)&spawn_id_bad,	0},
+{"expect_after",exp_proc(Exp_ExpectGlobalCmd),(ClientData)&exp_cmds[EXP_CMD_AFTER],0},
+{"expect_before",exp_proc(Exp_ExpectGlobalCmd),(ClientData)&exp_cmds[EXP_CMD_BEFORE],0},
+{"expect_user",	exp_proc(Exp_ExpectCmd),	(ClientData)&spawn_id_user,	0},
+{"expect_tty",	exp_proc(Exp_ExpectCmd),	(ClientData)&exp_dev_tty,	0},
+{"expect_background",exp_proc(Exp_ExpectGlobalCmd),(ClientData)&exp_cmds[EXP_CMD_BG],0},
+{"match_max",	exp_proc(Exp_MatchMaxCmd),	0,	0},
+{"remove_nulls",exp_proc(Exp_RemoveNullsCmd),	0,	0},
+{"parity",	exp_proc(Exp_ParityCmd),		0,	0},
+{"timestamp",	exp_proc(Exp_TimestampCmd),	0,	0},
 {0}};
 
 void
@@ -3038,6 +2996,7 @@ Tcl_Interp *interp;
 	exp_cmd_init(&exp_cmds[EXP_CMD_BEFORE],EXP_CMD_BEFORE,EXP_PERMANENT);
 	exp_cmd_init(&exp_cmds[EXP_CMD_AFTER ],EXP_CMD_AFTER, EXP_PERMANENT);
 	exp_cmd_init(&exp_cmds[EXP_CMD_BG    ],EXP_CMD_BG,    EXP_PERMANENT);
+	exp_cmd_init(&exp_cmds[EXP_CMD_FG    ],EXP_CMD_FG,    EXP_TEMPORARY);
 
 	/* preallocate to one element, so future realloc's work */
 	exp_cmds[EXP_CMD_BEFORE].ecd.cases = 0;
