@@ -16,12 +16,41 @@ would appreciate credit if you use this file or parts of it.
 
 #include "expect_cf.h"
 
+/*
+   The following functions are linked from the Tcl library.  They
+   don't cause anything else in the library to be dragged in, so it
+   shouldn't cause any problems (e.g., bloat).
+
+   The functions are relatively small but painful enough that I don't care
+   to recode them.  You may, if you absolutely want to get rid of any
+   vestiges of Tcl.
+*/
+extern char *TclGetRegError();
+
+
+
+#if defined(HAVE_PTYM) && defined(HAVE_PTMX)
+/*
+ * HP-UX 10.0 with streams (optional) have both PTMX and PTYM.  I don't
+ * know which is preferred but seeing as how the HP trap stuff is so
+ * unusual, it is probably safer to stick with the native HP pty support,
+ * too.
+ */
+#undef HAVE_PTMX
+#endif
+
 #ifdef HAVE_UNISTD_H
 #  include <unistd.h>
 #endif
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#ifdef NO_STDLIB_H
+#include "../compat/stdlib.h"
+#else
+#include <stdlib.h>
+#endif
+
 #ifdef HAVE_SYSMACROS_H
 #include <sys/sysmacros.h>
 #endif
@@ -38,6 +67,10 @@ would appreciate credit if you use this file or parts of it.
 #  include <fcntl.h>
 #endif
 
+#if defined(_SEQUENT_)
+#  include <sys/strpty.h>
+#endif
+
 #ifdef HAVE_PTMX
 #  include <sys/stropts.h>
 #endif
@@ -50,7 +83,7 @@ would appreciate credit if you use this file or parts of it.
 
 void debuglog();
 
-extern int errno;
+#include <errno.h>
 /*extern char *sys_errlist[];*/
 
 #ifndef TRUE
@@ -70,9 +103,14 @@ extern int errno;
 extern char *getpty();
 static char *master_name;
 static char slave_name[] = "/dev/ptyXX";
+static char	*tty_bank;		/* ptr to char [p-z] denoting
+					   which bank it is */
+static char	*tty_num;		/* ptr to char [0-f] denoting
+					   which number it is */
 #endif
 
-#if defined(_SEQUENT_)
+#if defined(_SEQUENT_) && !defined(HAVE_PTMX)
+/* old-style SEQUENT, new-style uses ptmx */
 static char *master_name, *slave_name;
 #endif /* _SEQUENT */
 
@@ -99,7 +137,7 @@ static char master_name[MAXPTYNAMELEN];
 static char slave_name[MAXPTYNAMELEN];
 #endif
 
-#if !defined(HAVE_GETPTY) && !defined(HAVE__GETPTY) && !defined(HAVE_PTC) && !defined(HAVE_PTC_PTS) && !defined(HAVE_PTMX) && !defined(HAVE_CONVEX_GETPTY) && !defined(_SEQUENT_)
+#if !defined(HAVE_GETPTY) && !defined(HAVE__GETPTY) && !defined(HAVE_PTC) && !defined(HAVE_PTC_PTS) && !defined(HAVE_PTMX) && !defined(HAVE_CONVEX_GETPTY) && !defined(_SEQUENT_) && !defined(HAVE_SCO_CLIST_PTYS) && !defined(HAVE_OPENPTY)
 #ifdef HAVE_PTYM
 			/* strange order and missing d is intentional */
 static char	banks[] = "pqrstuvwxyzabcefghijklo";
@@ -111,7 +149,7 @@ static char	*slave_num;
 static char	banks[] = "pqrstuvwxyzPQRSTUVWXYZ";
 static char	master_name[] = "/dev/ptyXX";
 static char	slave_name [] = "/dev/ttyXX";
-#endif
+#endif /* HAVE_PTYM */
 
 static char	*tty_type;		/* ptr to char [pt] denoting
 					   whether it is a pty or tty */
@@ -121,10 +159,19 @@ static char	*tty_num;		/* ptr to char [0-f] denoting
 					   which number it is */
 #endif
 
+#if defined(HAVE_SCO_CLIST_PTYS)
+#  define MAXPTYNAMELEN 64
+static char master_name[MAXPTYNAMELEN];
+static char slave_name[MAXPTYNAMELEN];
+#endif /* HAVE_SCO_CLIST_PTYS */
+
+#ifdef HAVE_OPENPTY
+static char master_name[64];
+static char slave_name[64];
+#endif
 
 char *exp_pty_slave_name;
-
-
+char *exp_pty_error;
 
 #if 0
 static void
@@ -272,7 +319,7 @@ char *s;	/* stty args */
 void
 exp_init_pty()
 {
-#if !defined(HAVE_GETPTY) && !defined(HAVE__GETPTY) && !defined(HAVE_PTC) && !defined(HAVE_PTC_PTS) && !defined(HAVE_PTMX) && !defined(HAVE_CONVEX_GETPTY) && !defined(_SEQUENT_)
+#if !defined(HAVE_GETPTY) && !defined(HAVE__GETPTY) && !defined(HAVE_PTC) && !defined(HAVE_PTC_PTS) && !defined(HAVE_PTMX) && !defined(HAVE_CONVEX_GETPTY) && !defined(_SEQUENT_) && !defined(HAVE_SCO_CLIST_PTYS) && !defined(HAVE_OPENPTY)
 #ifdef HAVE_PTYM
 	static char dummy;
 	tty_bank =  &master_name[strlen("/dev/ptym/pty")];
@@ -305,9 +352,15 @@ getptymaster()
 	char *hex, *bank;
 	struct stat stat_buf;
 	int master = -1;
+	int slave = -1;
 	int num;
 
+	exp_pty_error = 0;
+
+#define TEST_PTY 1
+
 #if defined(HAVE_PTMX) || defined(HAVE_PTMX_BSD)
+#undef TEST_PTY
 #if defined(HAVE_PTMX_BSD)
         if ((master = open("/dev/ptmx_bsd", O_RDWR)) == -1) return(-1);
 #else
@@ -326,6 +379,7 @@ getptymaster()
 #endif
 
 #if defined(HAVE__GETPTY)		/* SGI needs it this way */
+#undef TEST_PTY
 	slave_name = _getpty(&master, O_RDWR, 0600, 0);
 	if (slave_name == NULL)
 		return (-1);	
@@ -334,6 +388,7 @@ getptymaster()
 #endif
 
 #if defined(HAVE_PTC) && !defined(HAVE__GETPTY)	/* old SGI, version 3 */
+#undef TEST_PTY
 	master = open("/dev/ptc", O_RDWR);
 	if (master >= 0) {
 		int ptynum;
@@ -350,6 +405,7 @@ getptymaster()
 #endif
 
 #if defined(HAVE_GETPTY) && !defined(HAVE__GETPTY)
+#undef TEST_PTY
 	master = getpty(master_name, slave_name, O_RDWR);
 	/* is it really necessary to verify slave side is usable? */
 	exp_pty_slave_name = slave_name;
@@ -357,6 +413,7 @@ getptymaster()
 #endif
 
 #if defined(HAVE_PTC_PTS)
+#undef TEST_PTY
 	master = open("/dev/ptc",O_RDWR);
 	if (master >= 0) {
 		/* never fails */
@@ -366,17 +423,34 @@ getptymaster()
 	return(master);
 #endif
 
-#if defined(_SEQUENT_)
+#if defined(_SEQUENT_) && !defined(HAVE_PTMX)
+#undef TEST_PTY
 	/* old-style SEQUENT, new-style uses ptmx */
 	master = getpseudotty(&slave_name, &master_name);
 	exp_pty_slave_name = slave_name;
 	return(master);
 #endif /* _SEQUENT_ */
 
-	/* all pty allocation mechanisms after this require testing */
+#if defined(HAVE_OPENPTY)
+#undef TEST_PTY
+	if (openpty(&master, &slave, master_name, 0, 0) != 0) {
+		close(master);
+		close(slave);
+		return -1;
+	}
+	strcpy(slave_name, ttyname(slave));
+	exp_pty_slave_name = slave_name;
+	close(slave);
+	return master;
+#endif /* HAVE_OPENPTY */
+
+#if defined(TEST_PTY)
+	/*
+	 * all pty allocation mechanisms after this require testing
+	 */
 	if (exp_pty_test_start() == -1) return -1;
 
-#if !defined(HAVE_GETPTY) && !defined(HAVE__GETPTY) && !defined(HAVE_PTC) && !defined(HAVE_PTC_PTS) && !defined(HAVE_PTMX) && !defined(HAVE_CONVEX_GETPTY) && !defined(HAVE_PTYM) && !defined(_SEQUENT_)
+#if !defined(HAVE_CONVEX_GETPTY) && !defined(HAVE_PTYM) && !defined(HAVE_SCO_CLIST_PTYS)
 	for (bank = banks;*bank;bank++) {
 		*tty_bank = *bank;
 		*tty_num = '0';
@@ -391,6 +465,22 @@ getptymaster()
 	}
 #endif
 
+#ifdef HAVE_SCO_CLIST_PTYS
+        for (num = 0; ; num++) {
+            char num_str [16];
+
+            sprintf (num_str, "%d", num);
+            sprintf (master_name, "%s%s", "/dev/ptyp", num_str);
+            if (stat (master_name, &stat_buf) < 0)
+                break;
+            sprintf (slave_name, "%s%s", "/dev/ttyp", num_str);
+
+            master = exp_pty_test (master_name, slave_name, 0, num_str);
+            if (master >= 0)
+                goto done;
+        }
+#endif
+
 #ifdef HAVE_PTYM
 	/* systems with PTYM follow this idea:
 
@@ -479,12 +569,12 @@ getptymaster()
 	}
 #endif
 
-#if !defined(HAVE_GETPTY) && !defined(HAVE__GETPTY) && !defined(HAVE_PTC) && !defined(HAVE_PTC_PTS) && !defined(HAVE_PTMX)
  done:
 	exp_pty_test_end();
 	exp_pty_slave_name = slave_name;
 	return(master);
-#endif
+
+#endif /* defined(TEST_PTY) */
 }
 
 /* if slave is opened in a child, slave_control(1) must be executed after */
@@ -516,18 +606,18 @@ char *stty_args;
 #if defined(HAVE_PTMX_BSD)
 	if (ioctl (slave, I_LOOK, buf) != 0)
 		if (ioctl (slave, I_PUSH, "ldterm")) {
-			debuglog("ioctl(%s,I_PUSH,\"ldterm\") = %s\n",strerror(errno));
+			debuglog("ioctl(%s,I_PUSH,\"ldterm\") = %s\n",Tcl_ErrnoMsg(errno));
 	}
 #else
 #if defined(HAVE_PTMX)
 	if (ioctl(slave, I_PUSH, "ptem")) {
-		debuglog("ioctl(%s,I_PUSH,\"ptem\") = %s\n",strerror(errno));
+		debuglog("ioctl(%s,I_PUSH,\"ptem\") = %s\n",Tcl_ErrnoMsg(errno));
 	}
 	if (ioctl(slave, I_PUSH, "ldterm")) {
-		debuglog("ioctl(%s,I_PUSH,\"ldterm\") = %s\n",strerror(errno));
+		debuglog("ioctl(%s,I_PUSH,\"ldterm\") = %s\n",Tcl_ErrnoMsg(errno));
 	}
 	if (ioctl(slave, I_PUSH, "ttcompat")) {
-		debuglog("ioctl(%s,I_PUSH,\"ttcompat\") = %s\n",strerror(errno));
+		debuglog("ioctl(%s,I_PUSH,\"ttcompat\") = %s\n",Tcl_ErrnoMsg(errno));
 	}
 #endif
 #endif
@@ -641,3 +731,9 @@ int fd;
 	return(found);
 }
 #endif
+
+void
+exp_pty_exit()
+{
+	/* a stub so we can do weird things on the cray */
+}
